/*+*****************************************************************************
*                                                                              *
* File: input.c                                                                *
*                                                                              *
* Description: input handling, i.e. where to listen to with XtAddInput         *
*                                                                              *
**-****************************************************************************/

#ifndef __lint
char input_vers[] = "@(#)input.c	1.2	01/07/98	Written by Lionel Cons";
#endif /* __lint */

/******* Include Files ********************************************************/

#include <X11/Xlib.h>

#include "input.h"
#include "conn.h"
#include "xu.h"

/******* Constants ************************************************************/

#define NO_INTERVAL_ID   (XtIntervalId)0
#define IDLE_INTERVAL_ID XtAppAddTimeOut(input_app_context, 1000, input_idle_tasks, (XtPointer)0)

/******* Macros ***************************************************************/

/******* External Stuff *******************************************************/

/******* Local Stuff **********************************************************/

static XtAppContext input_app_context;
static void (*input_destroy_connection)(CONN *);
static XtIntervalId input_idle_id = NO_INTERVAL_ID;

/*
 * what to do from time to time
 */
/*ARGSUSED*/
static void input_idle_tasks(XtPointer data, XtIntervalId *id)
{
  CONN *cop;
  int todo;

  todo = conn_idle(&cop);
#ifdef DEBUG
  if (DEBUGGING(DBG_DISPATCH))
    dprintf(NULL, "idle tasks activates %s, %s",
	    (cop  ? cop->id    : "nothing"),
	    (todo ? "still on" : "now off"));
#endif
  if (cop) {
      /* re-activate this connection */
      input_update_connection(cop);
  }
  /* call me again? */
  if (todo)
    input_idle_id = IDLE_INTERVAL_ID;
  else
    input_idle_id = NO_INTERVAL_ID;
}

/*
 * check and update the input ids
 */
static void input_check(XtInputId *id, int fd, XtInputMask cond,
			XtInputCallbackProc proc, XtPointer data)
{
  switch (cond) {
  case XtInputReadMask:
  case XtInputWriteMask:
    if (*id != NO_INPUT_ID) return;
    *id = XtAppAddInput(input_app_context, fd, (XtPointer)cond, proc, data);
#ifdef DEBUG
    if (DEBUGGING(DBG_DISPATCH)) {
      CONN *c = (CONN *)data;
      dprintf(c->tag, "new %s input id",
	      ((id == &(c->ui.cid.in)) ? "client read " :
	       ((id == &(c->ui.cid.out)) ? "client write" :
		((id == &(c->ui.sid.in)) ? "server read " :
		 ((id == &(c->ui.sid.out)) ? "server write" : "???")))));
    }
#endif
    break;
  case XtInputNoneMask:
    if (*id == NO_INPUT_ID) return;
    XtRemoveInput(*id);
    *id = NO_INPUT_ID;
#ifdef DEBUG
    if (DEBUGGING(DBG_DISPATCH)) {
      CONN *c = (CONN *)data;
      dprintf(c->tag, "no  %s input id",
	      ((id == &(c->ui.cid.in)) ? "client read " :
	       ((id == &(c->ui.cid.out)) ? "client write" :
		((id == &(c->ui.sid.in)) ? "server read " :
		 ((id == &(c->ui.sid.out)) ? "server write" : "???")))));
    }
#endif
    break;
  default:
    die("bad condition in input_check: %d", cond);
  }
}

/*
 * freeze a connection
 */
static void input_freeze(CONN *c)
{
  XtPointer data = (XtPointer)c;
  IO *iop;
  ID *idp;
  
  /* client side */
  iop = &c->client;
  idp = &c->ui.cid;
  input_check(&idp->in,  iop->fd_in,  XtInputNoneMask, NULL, data);
  input_check(&idp->out, iop->fd_out, XtInputNoneMask, NULL, data);
  /* server side */
  iop = &c->server;
  idp = &c->ui.sid;
  input_check(&idp->in,  iop->fd_in,  XtInputNoneMask, NULL, data);
  input_check(&idp->out, iop->fd_out, XtInputNoneMask, NULL, data);
}

/*
 * update the input ids and other things
 */
static void input_update(CONN *c, IO *iop, ID *idp, XtInputCallbackProc proc)
{
  XtPointer data = (XtPointer)c;

  /* frozen connection or dialog shown: don't do I/O */
  if (BITTST(c->flags, CF_FROZEN) || c->ui.alert) {
    input_freeze(c);
    return;
  }

  /* an alert has just been cleared, we have to trigger proc */
  if (c->xinfo.alert) {
    (*proc)(c, NULL, NULL);
    return;
  }

  /* error handling */
  if (BITTST(iop->flags, IO_ERROR)) {
    input_destroy_connection(c);
    return;
  }

  /* normal case */
  if (BITTST(iop->flags, IO_READ))
    input_check(&idp->in,  iop->fd_in,  XtInputReadMask,  proc, data);
  else
    input_check(&idp->in,  iop->fd_in,  XtInputNoneMask,  proc, data);
  if (BITTST(iop->flags, IO_WRITE))
    input_check(&idp->out, iop->fd_out, XtInputWriteMask, proc, data);
  else
    input_check(&idp->out, iop->fd_out, XtInputNoneMask,  proc, data);
}

/*
 * callback to dismiss an alert
 */
static void input_alert_dismiss_cb(Widget w, CONN *c)
{
  xu_dismiss(w, NULL);
  c->ui.alert = (Widget)NULL;
  /* update the input */
  input_update_connection(c);
}

/*
 * callback to replace next packet by a NoOp
 */
static void input_alert_noop_cb(Widget w, CONN *c)
{
  BITSET(c->flags, CF_NOOP);
  input_alert_dismiss_cb(w, c);
}

/*
 * callback to kill a connection
 */
static void input_alert_kill_cb(Widget w, CONN *c)
{
  xu_dismiss(w, NULL);
  c->ui.alert = (Widget)NULL;
  input_destroy_connection(c);
}

/*
 * popup an alert
 */
static void input_alert(CONN *c)
{
  ANSWER allow, killit, noop;
  char scratch[256], *cp;

  BITCLR(c->flags, CF_NOOP);
  allow.text = "Allow";
  allow.callback = (XtCallbackProc)input_alert_dismiss_cb;
  allow.data = (XtPointer)c;
  killit.text = "Kill";
  killit.callback = (XtCallbackProc)input_alert_kill_cb;
  killit.data = (XtPointer)c;
  noop.text = "NoOp";
  noop.callback = (XtCallbackProc)input_alert_noop_cb;
  noop.data = (XtPointer)c;
  cp = scratch;
  sprintf(cp, "A %s has been detected on connection %s from %s:\n",
	  (c->xinfo.alert_level == AL_DUBIOUS ? "dubious X request" : "security violation"),
	  c->id, c->name);
  cp += strlen(cp);
  sprintf(cp, "     %s\nWhat do you want to do?", c->xinfo.alert);
  c->ui.alert = xu_ask(scratch, &allow, &killit, &noop);
}

/*
 * what to do when the client side wants to do I/O
 */
/*ARGSUSED*/
static void input_server(CONN *c, int *source, XtInputId *id)
{
  /* frozen connection or dialog shown: don't do I/O! */
  if (BITTST(c->flags, CF_FROZEN) || c->ui.alert) return;

  /* do some work */
  conn_server_read(c);
  conn_server_write(c);

  /* update input */
  input_update(c, &c->server, &c->ui.sid, (XtInputCallbackProc)input_server);
}

/*
 * what to do when the server side wants to do I/O
 */
/*ARGSUSED*/
static void input_client(CONN *c, int *source, XtInputId *id)
{
  /* frozen connection or dialog shown: don't do I/O! */
  if (BITTST(c->flags, CF_FROZEN) || c->ui.alert) return;

  /* do some work */
  conn_client_read(c);
  conn_client_write(c);

  /* alert detected? */
  if (c->xinfo.alert) input_alert(c);

  /* fake packets to queue? */
  if (c->xinfo.fake_packet && input_idle_id == NO_INTERVAL_ID)
    input_idle_id = IDLE_INTERVAL_ID;

  /* update input */
  input_update(c, &c->client, &c->ui.cid, (XtInputCallbackProc)input_client);
}

/*
 * update all the input callbacks for a connection
 */
void input_update_connection(CONN *c)
{
  input_update(c, &c->client, &c->ui.cid, (XtInputCallbackProc)input_client);
  input_update(c, &c->server, &c->ui.sid, (XtInputCallbackProc)input_server);
}

/*
 * setup
 */
void input_setup(XtAppContext ac, void (*dc)(CONN *))
{
  input_app_context = ac;
  input_destroy_connection = dc;
}
