/*+*****************************************************************************
*                                                                              *
* File: xs.c                                                                   *
*                                                                              *
* Description: X sockets handling                                              *
*                                                                              *
**-****************************************************************************/

#ifndef __lint
char xs_vers[] = "@(#)xs.c	1.7	01/23/98	Written by Lionel Cons";
#endif /* __lint */

/******* Include Files ********************************************************/

#include "xs.h"
#include "util.h"
#include "host.h"
#include "auth.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/stat.h>

/******* Constants ************************************************************/

/*
 * from Xproto.h
 */
#define X_TCP_PORT 6000		/* add display number */

/*
 * dirty hack for HP-UX :-(, see Xtranssock.c
 */
#ifdef __hpux
#ifdef HPUX10
#define X_UNIX_PATH	"/var/spool/sockets/X11/"
#define X_UNIX_DIR	"/var/spool/sockets/X11"
#else
#define X_UNIX_PATH	"/usr/spool/sockets/X11/"
#define X_UNIX_DIR	"/usr/spool/sockets/X11"
#endif /* __hpux */
#else
#define X_UNIX_PATH	"/tmp/.X11-unix/X"
#define X_UNIX_DIR	"/tmp/.X11-unix"
#endif

/*
 * length of the queue for pending connections
 */
#define LISTEN_BACKLOG	10

/******* Macros ***************************************************************/

/******* External Stuff *******************************************************/

/******* Local Stuff **********************************************************/

static HOST *xs_host;		/* host of the X server */
static int   xs_dispno;		/* display number of the X server */

static int   xs_fd = -1;	/* fd of the INET incomming socket */
static int   xs_fdl = -1;	/* fd of the UNIX incomming socket */
static char  xs_un_path[128];	/* path of the UNIX incomming socket */

/*
 * make a TCP connection to the X server
 * from Xtranssock.c/TRANS(SocketINETConnect)
 */
int xs_connect(void)
{
  struct sockaddr_in s_in;
  int fd;

  memset((char *)&s_in, '\0', sizeof(s_in));
  s_in.sin_family = AF_INET;
  s_in.sin_addr = xs_host->addr;
  s_in.sin_port = htons(X_TCP_PORT + xs_dispno);
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "can't create INET socket %s:%d: %s",
	      xs_host->ip_name, X_TCP_PORT + xs_dispno, ERROR);
#endif
    return(-1);
  }
#ifdef TCP_NODELAY
  {
    int one = 1;
    setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));
  }
#endif
  if (connect(fd, (struct sockaddr *)&s_in, sizeof(s_in)) < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "can't connect INET socket %s:%d: %s",
	      xs_host->ip_name, X_TCP_PORT + xs_dispno, ERROR);
#endif
    return(-1);
  }
  return(fd);
}

/*
 * make a UNIX socket connection to the X server
 * from Xtranssock.c/TRANS(SocketUNIXConnect)
 */
int xs_connect_local(void)
{
  struct sockaddr_un s_un;
  int fd;

  memset((char *)&s_un, '\0', sizeof(s_un));
  s_un.sun_family = AF_UNIX;
  sprintf(s_un.sun_path, "%s%d", X_UNIX_PATH, xs_dispno);
  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (fd < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "can't create UNIX socket %s: %s", s_un.sun_path, ERROR);
#endif
    return(-1);
  }
  if (connect(fd, (struct sockaddr *)&s_un, sizeof(s_un)) < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "can't connect UNIX socket %s: %s", s_un.sun_path, ERROR);
#endif
    return(-1);
  }
  return(fd);
}

/*
 * accept a TCP connection
 */
int xs_accept(HOST **host)
{
  int newfd;
  struct sockaddr_in s_in;
  int sinlen;
  HOST *newhost;

  /* accept */
  sinlen = sizeof(s_in);
  newfd = accept(xs_fd, (void *)&s_in, &sinlen);
  if (newfd < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "accept failed: %s", ERROR);
#endif
    return(-1);
  }

  /* check host */
  newhost = host_create_by_addr(&s_in.sin_addr);
  if (!newhost) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "accept rejected for %s: unknown host",
	      inet_ntoa(s_in.sin_addr));
#endif
    return(-1);
  }

  /* return */
#ifdef DEBUG
  if (DEBUGGING(DBG_CONN))
    dprintf(NULL, "detected connection from %s [%s], port %d",
	      newhost->name, newhost->ip_name, ntohs(s_in.sin_port));
#endif
  *host = newhost;
  return(newfd);
}

/*
 * accept a UNIX socket connection
 */
int xs_accept_local(void)
{
  int newfd;
  struct sockaddr_un s_un;
  int sunlen;

  /* accept */
  sunlen = sizeof(s_un);
  newfd = accept(xs_fdl, (void *)&s_un, &sunlen);
  if (newfd < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR))
      dprintf(NULL, "local accept failed: %s", ERROR);
#endif
    return(-1);
  }

  /* return */
#ifdef DEBUG
  if (DEBUGGING(DBG_CONN))
    dprintf(NULL, "detected local connection");
#endif
  return(newfd);
}

/*
 * create the sockets to act as a fake X server
 * see connection.c and Xtranssock.c
 */
static void xs_listen(int *dispno_ret, int hunt)
{
  int dispno = *dispno_ret;
  struct sockaddr_in s_in;
  struct sockaddr_un s_un;
  int res, oldumask;

top:
  /* INET socket */
  memset((char *)&s_in, '\0', sizeof(s_in));
  s_in.sin_family = AF_INET;
  s_in.sin_addr.s_addr = htonl(INADDR_ANY);
  s_in.sin_port = htons(X_TCP_PORT + dispno);
  xs_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (xs_fd < 0) die("can't create INET socket: %s", ERROR);
  /* UNIX socket */
  oldumask = umask(0);
  if (mkdir(X_UNIX_DIR, 0777) == 0) chmod(X_UNIX_DIR, 0777);
  memset((char *)&s_un, '\0', sizeof(s_un));
  s_un.sun_family = AF_UNIX;
  sprintf(xs_un_path, "%s%d", X_UNIX_PATH, dispno);
  strcpy(s_un.sun_path, xs_un_path);
  xs_fdl = socket(AF_UNIX, SOCK_STREAM, 0);
  if (xs_fdl < 0) die("can't create UNIX socket: %s", ERROR);
#ifdef SO_REUSEADDR
  {
    int one = 1;
    res = setsockopt(xs_fd, SOL_SOCKET, SO_REUSEADDR, (void *)&one, sizeof(one));
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR) && res < 0)
      dprintf(NULL, "can't setsockopt SO_REUSEADDR: %s", ERROR);
#endif
  }
#endif /* SO_REUSEADDR */
  /* try to bind INET socket */
  res = bind(xs_fd, (void *)&s_in, sizeof(s_in));
  if (res < 0) {
    if ((errno == EADDRINUSE) && hunt) {
      dispno++;
      close(xs_fd);
      close(xs_fdl);
      goto top;
    }
    die("can't bind INET socket: %s", ERROR);
  }
  /* try to bind UNIX socket */
  unlink(xs_un_path);
  res = bind(xs_fdl, (void *)&s_un, sizeof(s_un));
  if (res < 0) {
    if ((errno == EADDRINUSE) && hunt) {
      dispno++;
      close(xs_fd);
      close(xs_fdl);
      goto top;
    }
    die("can't bind UNIX socket: %s", ERROR);
  }
  /* additional socket options */
#ifdef SO_LINGER
  {
    static int linger[2] = { 0, 0 };
    res = setsockopt(xs_fd, SOL_SOCKET, SO_LINGER, (void *)linger, sizeof(linger));
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR) && res < 0)
      dprintf(NULL, "can't setsockopt SO_LINGER: %s", ERROR);
#endif
  }
#else /* SO_LINGER */
#ifdef SO_DONTLINGER
  {
    res = setsockopt(xs_fd, SOL_SOCKET, SO_DONTLINGER, (void *)NULL, 0);
#ifdef DEBUG
    if (DEBUGGING(DBG_ERROR) && res < 0)
      dprintf(NULL, "can't setsockopt SO_DONTLINGER: %s", ERROR);
#endif
  }
#endif /* SO_DONTLINGER */
#endif /* SO_LINGER */
  /* try to listen INET socket */
  res = listen(xs_fd,  LISTEN_BACKLOG);
  if (res < 0) die("can't listen to INET socket: %s", ERROR);
  /* try to listen UNIX socket */
  res = listen(xs_fdl, LISTEN_BACKLOG);
  if (res < 0) die("can't listen to UNIX socket: %s", ERROR);

  /* return the new display number */
  umask(oldumask);
  *dispno_ret = dispno;
}

/*
 * parse the DISPLAY
 */
static void parse_display(void)
{
  char *dv, *cp, *dn;

  /* find the current DISPLAY */
  dv = getenv("DISPLAY");
  if (dv == 0) die("the DISPLAY is not set");

  /* analyze it, algorithm from XConnDis.c/_XConnectDisplay */
  for (cp=dv; *cp && (*cp != ':'); cp++);
  if (!*cp) die("bad DISPLAY (missing colon): %s", dv);
  if (cp[1] == ':') die("bad DISPLAY (DECnet is not supported): %s", dv);
  *cp++ = '\0';
  /* end of host part */
  dn = cp;
  for (cp=dn; *cp && (*cp != '.'); cp++);
  *cp = '\0';
  /* end of number part */
  xs_dispno = atoi(dn);
  if (*dv == '\0' || !strcmp(dv, "unix") || !strcmp(dv, "localhost")) {
    /* local DISPLAY */
    xs_host = localhost;
  } else {
    /* remote DISPLAY */
    xs_host = host_create_by_name(dv);
    if (xs_host == NULL) die("bad DISPLAY (unknown host): %s", dv);
  }
#ifdef DEBUG
  if (DEBUGGING(DBG_MISC))
    dprintf(NULL, "using %s X server %d on %s",
	    (xs_host == localhost ? "local" : "remote"),
	    xs_dispno, xs_host->name);
#endif
}

/*
 * create a fake X server and initialise some data in order
 * to be able to connect to the real X server when needed
 */
void xs_setup(int *dispno_ret, int hunt, int *fd_ret, int *fdl_ret)
{
  parse_display();
  auth_setup(xs_host, xs_dispno);
  xs_listen(dispno_ret, hunt);
  *fd_ret  = xs_fd;
  *fdl_ret = xs_fdl;
#ifdef DEBUG
  if (DEBUGGING(DBG_MISC))
    dprintf(NULL, "created sockets for dispno %d: INET=%d UNIX=%d",
	    *dispno_ret, xs_fd, xs_fdl);
#endif
}

/*
 * close the fake X server
 */
void xs_close(void)
{
  if (xs_fd > 0)  close(xs_fd);
  if (xs_fdl > 0) close(xs_fdl);
  unlink(xs_un_path); /* so that it can be reused... */
}
